/** Apply forward feeding to a fully connect neural network.
 * This struct stores the final output as well as the activations that occur
 * for use in backpropagation
 *
 */
#include "activation_function.hpp"
#include "matrix.hpp"
#include <vector>

template <class ActivationFunction> struct ForwardFeed {
  std::vector<Matrix<float>> m_activations;
  std::vector<float> m_yhat;

  ForwardFeed(const std::vector<float> &x,
              const std::vector<Matrix<float>> &weights) {
    // Convert input vector to matrix
    Matrix<float> A = Matrix<float>(x.size(), 1, x);

    // Feed each layer forward except the last layer using the user specified
    // activation function
    m_activations.reserve(weights.size());
    for (size_t i = 0; i < weights.size() - 1; i++) {
      // Calculate Z = W * A
      Matrix Z = weights[i] * A;

      // Apply activation function
      ActivationFunction::apply(Z.data());
      m_activations.push_back(A);
      A = Z;
    }

    // Always use soft max for the final layer
    Matrix Z = weights.back() * A;
    SoftMax::apply(Z.data());

    m_yhat = Z.data();
  };
};
